"""Nodes of the abstract syntax tree of the program:

- Uop/Bop represent unary/binary arithmetic operator
- ArithExpr and its children describe arithmetic expressions
- BoolExpr and its children describe boolean expressions
- Instr and its children describe instruction

"""
from enum import Enum
from dataclasses import dataclass

class Uop(Enum):
    """unary operators. Currently only unary minus"""
    OPP = 1
    def __str__(self):
        if self is Uop.OPP:
            return "-"
        return ""

class Bop(Enum):
    """binary operators: addition, multiplication, and division"""
    ADD = 1
    MUL = 2
    DIV = 3
    def __str__(self):
        match self:
            case Bop.ADD: return "+"
            case Bop.MUL: return "*"
            case Bop.DIV: return "/"

class Expr:
    """generic class for expressions, either arithmetic or boolean"""
    pass

@dataclass
class ArithExpr(Expr):
    """generic class for arithmetic expressions"""
    pass

@dataclass
class AECst(ArithExpr):
    """integer constants"""
    value: int
    def __str__(self):
        return f"{self.value}"

@dataclass
class AEVar(ArithExpr):
    """variables"""
    var: str
    def __str__(self):
        return self.var

@dataclass
class AEUop(ArithExpr):
    """unary operation"""
    uop: Uop
    expr: ArithExpr
    def __str__(self):
        return f"{self.uop}{self.expr}"

@dataclass
class AEBop(ArithExpr):
    """binary operation"""
    bop: Bop
    left_expr: ArithExpr
    right_expr: ArithExpr
    def __str__(self):
        #TODO: only put parentheses if necessary
        if isinstance(self.left_expr, AEBop) or isinstance(self.left_expr, AEBop):
            return f"({self.left_expr}) {self.bop} ({self.right_expr})"
        elif isinstance(self.left_expr, AEBop):
            return f"{self.left_expr} {self.bop} ({self.right_expr})"
        elif isinstance(self.right_expr, AEBop):
            return f"({self.left_expr}) {self.bop} {self.right_expr}"
        else:
            return f"{self.left_expr} {self.bop} {self.right_expr}"


@dataclass
class BoolExpr(Expr):
    """generic class for boolean expressions"""
    pass

@dataclass
class BEPlain(BoolExpr):
    """arithmetic expression (true if not 0)"""
    aexpr: ArithExpr
    def __str__(self):
        return f"{self.aexpr}"

@dataclass
class BEEq(BoolExpr):
    """equality between two arithmetic expression"""
    left_expr: ArithExpr
    right_expr:ArithExpr
    def __str__(self):
        return f"{self.left_expr} = {self.right_expr}"

@dataclass
class BELeq(BoolExpr):
    """less or equal"""
    left_expr: ArithExpr
    right_expr: ArithExpr
    def __str__(self):
        return f"{self.left_expr} <= {self.right_expr}"

@dataclass
class BENeg(BoolExpr):
    """Negation"""
    expr: BoolExpr
    def __str__(self):
        return f"!{self.expr}"

class Instr:
    """generic class for instruction"""
    pass

@dataclass
class ISkip(Instr):
    """Skip (do nothing)"""
    def __str__(self):
        return "skip"

@dataclass
class ISet(Instr):
    """assignment to a variable"""
    var: str
    expr: ArithExpr
    def __str__(self):
        return f"{self.var} := {self.expr}"

@dataclass
class ISeq(Instr):
    """sequence of two instructions"""
    first: Instr
    second: Instr
    def __str__(self):
        return f"{self.first};\n{self.second}"

@dataclass
class ICond(Instr):
    """if then else"""
    cond: BoolExpr
    true_branch: Instr
    false_branch: Instr
    def __str__(self):
        return f"if {self.cond} then\n{self.true_branch}\nelse\n{self.false_branch}\nfi"

@dataclass
class ILoop(Instr):
    """while loop"""
    cond: BoolExpr
    body: Instr
    def __str__(self):
        return f"while {self.cond} do\n{self.body}\ndone"

def variables_of_aexpr(e: ArithExpr) -> frozenset[str]:
    """return the set of variables appearing in the expression"""
    match e:
        case AECst(_): return frozenset()
        case AEVar(var): return frozenset(var)
        case AEUop(expr=e): return variables_of_aexpr(e)
        case AEBop(left_expr=le,right_expr=re):
            return variables_of_aexpr(le) | variables_of_aexpr(re)
        case _: assert False

def variables_of_bexpr(e: BoolExpr) -> frozenset[str]:
    """return the set of variable appearing in the expression"""
    match e:
        case BEPlain(expr): return variables_of_aexpr(expr)
        case BEEq(left_expr=le,right_expr=re):
            return variables_of_aexpr(le) | variables_of_aexpr(re)
        case BELeq(left_expr=le,right_expr=re):
            return variables_of_aexpr(le) | variables_of_aexpr(re)
        case BENeg(expr):
            return variables_of_bexpr(expr)
        case _: assert False

def variables_of_instr(instr: Instr) -> frozenset[str]:
    """return the set of variables appearing in a program"""
    match instr:
        case ISkip(): return frozenset()
        case ISet(var=v,expr=e):
            return frozenset([v]) | variables_of_aexpr(e)
        case ISeq(first=i1,second=i2):
            return variables_of_instr(i1) | variables_of_instr(i2)
        case ICond(cond=c,true_branch=tb,false_branch=fb):
            vbranches = variables_of_instr(tb) | variables_of_instr(fb)
            return variables_of_bexpr(c) | vbranches
        case ILoop(cond=c,body=b):
            return variables_of_bexpr(c) | variables_of_instr(b)
        case _ : assert False
